#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from typing import List, Dict

from PyQt5.QtCore import QProcessEnvironment, QDir

from astra_csp_cryptopro.cprocsp_process import CprocspProcess
from astra_csp_cryptopro.utils import Utils


class CertificateIndentity(dict):
    def __init__(self, identity_string: str) -> None:
        super().__init__()
        self.__identity_string = identity_string
        self.__dict_re = re.compile('(?P<key>.*?)=(?P<value>.*?)(,|$)')
        self.__setup_dict(self.__identity_string)

    def __str__(self) -> str:
        return self.__identity_string

    def __setup_dict(self, identity_string: str) -> None:
        for match in self.__dict_re.finditer(identity_string):
            if match:
                key = match.group('key').strip()
                value = match.group('value').strip()
                self[key] = value


class Certificate(dict):
    def __init__(self, fingerprint: str, subject: CertificateIndentity) -> None:
        super().__init__()
        self.__fingerprint = fingerprint
        self.__subject = subject

    def __repr__(self) -> str:
        return self.fingerprint + ' ' + str(self.subject)

    @property
    def fingerprint(self) -> str:
        return self.__fingerprint

    @property
    def subject(self) -> CertificateIndentity:
        return self.__subject


class CertificateProcess(CprocspProcess):
    def __init__(self, bin_dir: QDir) -> None:
        super().__init__(bin_dir=bin_dir)
        self.__setup()
        self.__fields_re = re.compile('^(?P<key>\\w.*?):(?P<value>.*?)$', re.MULTILINE | re.U)

    def __setup(self) -> None:
        self.set_bin_name('certmgr')
        self.setArguments(['-list'])
        env = QProcessEnvironment.systemEnvironment()
        env.insert('LC_ALL', 'C.UTF-8')
        self.setProcessEnvironment(env)

    def get_certificates(self) -> List[Certificate]:
        sertificates = []  # type: List[Certificate]
        output = Utils.to_utf8(self.standard_output_buffer)
        for certinfo in re.split('\\d+-------', output):
            fields = self.__extract_fields(certinfo)
            if 'Issuer' in fields and 'Subject' in fields:
                sha1_fields = [field for field in fields if 'SHA1' in field]
                if len(sha1_fields):
                    sha1_field = sha1_fields[0]
                    fields['SHA1'] = fields.pop(sha1_field)

                    issuer_id = CertificateIndentity(fields['Issuer'])
                    certificate_id = CertificateIndentity(fields['Subject'])
                    cert = Certificate(fingerprint=fields['SHA1'], subject=certificate_id)
                    cert.update(fields)
                    cert['Issuer'] = issuer_id
                    cert['Subject'] = certificate_id
                    sertificates.append(cert)
        return sertificates

    def __extract_fields(self, certificate: str) -> Dict[str, str]:
        fields = dict()  # type: Dict[str, str]
        for match in self.__fields_re.finditer(certificate):
            if match:
                key = match.group('key').strip()
                value = match.group('value').strip()
                fields[key] = value
        return fields

    def empty_list_error_occurred(self) -> bool:
        return self.output_error_code() == '0x8010002c'
