#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import subprocess
from typing import List, Optional, Tuple

from PyQt5.QtCore import QTemporaryDir
from PyQt5.QtWidgets import QApplication

from astra_csp_cryptopro.archivemount import ArchiveMount
from astra_csp_cryptopro.unpack_archive_thread import UnpackArchiveThread
from astra_csp_cryptopro.utils import Utils
from astra_csp_cryptopro.zip_archiver import ZipArchiver

translate = QApplication.translate


class ContainerException(Exception):
    def __init__(self, message: str, details: str = ''):
        super().__init__(message)
        self.__details = details

    @property
    def message(self) -> str:
        return str(self)

    @property
    def details(self) -> str:
        return self.__details


class UnmountError(ContainerException):
    pass


class Container:
    def __init__(self, filepath: str, tmpdir: str) -> None:
        self.__filepath = filepath
        self.__tmpdir = tmpdir
        self.__mountdir = QTemporaryDir(self.__tmpdir)
        self.__mountdir.setAutoRemove(False)

    @property
    def path(self) -> str:
        return self.__filepath

    @property
    def mount_dir(self) -> Optional[QTemporaryDir]:
        return self.__mountdir

    @property
    def name(self) -> str:
        return os.path.basename(self.__filepath)

    def files(self) -> List[str]:
        archive = ZipArchiver(self.__filepath, self.__mountdir.path())
        return [file_info.name for file_info in archive.info().files_info]

    def sources_path(self) -> str:
        return self.mount_dir.path()

    def create_unpack_thread(self, filenames: List[str] = None) -> UnpackArchiveThread:
        archive = ZipArchiver(self.__filepath, self.__mountdir.path())
        return UnpackArchiveThread(archive=archive, filenames=filenames)

    def unpacked(self) -> bool:
        if not self.__mountdir:
            return False

        if not self.__mountdir.isValid():
            return False

        archive = ZipArchiver(self.__filepath, self.__mountdir.path())
        return archive.validate_unpacked_files()

    def update(self) -> None:
        if self.__mountdir:
            archive = ZipArchiver(self.__filepath, self.__mountdir.path())
            archive.update()
            self.__mountdir.remove()

    def save(self, files: List[str], update: bool = False) -> None:
        tmpdir = QTemporaryDir(self.__tmpdir)

        for file in files:
            if file:
                temp_symlink_path = tmpdir.path() + '/' + os.path.basename(file) + '.temp'
                os.symlink(file, temp_symlink_path)
                os.rename(temp_symlink_path, tmpdir.path() + '/' + os.path.basename(file))

        archive = ZipArchiver(self.__filepath, tmpdir.path())
        archive.create(update)

    def contains(self, filename: str) -> bool:
        basename = os.path.basename(filename)
        archive = ZipArchiver(self.__filepath, self.__mountdir.path())
        equal_path = os.path.dirname(filename) == self.__mountdir.path()
        return Utils.file_is_exists(self.__filepath) and equal_path and archive.info().file_exists(basename)

    @classmethod
    def is_container(cls, filename: str) -> bool:
        return filename.endswith('.signed.zip')

    @classmethod
    def is_external_container(cls, filename: str) -> bool:
        dirname = os.path.dirname(filename)
        match = re.match(r'/run/user/\d+/fly-fm-vfs/archives/[0-9a-fA-F]+', dirname)
        return bool(match)

    @classmethod
    def list_archivemount_processes(cls) -> subprocess.CompletedProcess:
        uid = os.getuid()
        return subprocess.run("ps -u {uid} -o args | grep /usr/bin/archivemount".format(uid=uid),
                              stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                              shell=True, universal_newlines=True)

    @classmethod
    def find_archive_and_mountpoint(cls, filepath: str, output: str) -> Tuple[Optional[str], Optional[str]]:
        output_re = re.compile('/usr/bin/archivemount\\s+-o\\s+nobackup\\s+(?P<archive>.*)\\s+'
                               '(?P<point>/run/user/\\d+/fly-fm-vfs/archives/.*)')
        for match in output_re.finditer(output):
            if match:
                archive = match.group('archive')
                mountpoint = match.group('point')
                if archive == filepath or filepath.startswith(mountpoint):
                    return archive, mountpoint
        return None, None

    @classmethod
    def find_and_unmount(cls, path) -> None:
        result = Container.list_archivemount_processes()
        if result.returncode == 0:
            archive, mountpoint = Container.find_archive_and_mountpoint(path, result.stdout)
            if archive and mountpoint:
                Container.unmount(archive, mountpoint)

    @classmethod
    def find_external(cls, path: str) -> Tuple[Optional[str], Optional[str]]:
        result = Container.list_archivemount_processes()
        if result.returncode != 0:
            raise ContainerException(translate('text', "Failed to find archivemount processes"),
                                     translate('text', "ps utility exited with code {code}")
                                     .format(code=result.returncode))
        archive, mountpoint = Container.find_archive_and_mountpoint(path, result.stdout)
        if archive and mountpoint:
            return archive, mountpoint
        raise ContainerException(translate('text', "Failed to find archive and "
                                                   "mountpoint in archivemount processes data"))

    @classmethod
    def unmount(cls, archive: str, mountpoint: str) -> None:
        try:
            archivemount = ArchiveMount(archive, mountpoint)
            archivemount.unmount()
            Container.notify_flyfm()
        except Exception as exception:
            raise UnmountError(translate('text', "Failed to unmount container"), str(exception))

    @classmethod
    def notify_flyfm(cls) -> None:
        try:
            subprocess.run("/usr/bin/fly-fm --refresh", shell=True,
                           stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        except Exception:
            pass

    @classmethod
    def is_encrypted(cls, filepath: str) -> bool:
        try:
            return ZipArchiver.is_encrypted(filepath)
        except Exception as exception:
            raise ContainerException(translate('text', 'Failed to read container encryption status'), str(exception))

    @classmethod
    def check_encryption(cls, filepath: str) -> None:
        if Container.is_encrypted(filepath):
            raise ContainerException(translate('text', 'Container is encrypted.\n'
                                                       'Sorry, encrypted containers are not supported yet'))
