# -*- coding: utf-8 -*-

import os
import sys
from datetime import datetime
from enum import Enum
from typing import Optional, List, Tuple
from platform import machine

import psutil
from PyQt5 import uic
from PyQt5.QtCore import Qt, QTimer, QProcess, QDir, pyqtSignal, QFileInfo
from PyQt5.QtGui import QGuiApplication, QIcon, QColor, QPalette, QKeyEvent, QCloseEvent
from PyQt5.QtWidgets import QApplication, QMessageBox, QMainWindow, QFileDialog, QTabWidget, QLabel, \
    QPushButton, QTextBrowser, QComboBox, QCheckBox, QLineEdit, QStackedWidget

from astra_csp_cryptopro.application_cache import ApplicationCache
from astra_csp_cryptopro.certificate_process import Certificate
from astra_csp_cryptopro.certificate_process import CertificateProcess
from astra_csp_cryptopro.container import Container, ContainerException, UnmountError
from astra_csp_cryptopro.container_creation_thread import ContainerCreationThread
from astra_csp_cryptopro.cprocsp_process_dialog import CprocspProcessDialog
from astra_csp_cryptopro.create_file_signature_process import CreateFileSignatureProcess
from astra_csp_cryptopro.interactive_process_dialog import InteractiveProcessDialog
from astra_csp_cryptopro.merge_certificates_process import MergeCertificatesProcess
from astra_csp_cryptopro.package import Package
from astra_csp_cryptopro.pdf_conversion_process import PdfConversionProcess
from astra_csp_cryptopro.plugins.plugin_loader import PluginLoader
from astra_csp_cryptopro.process import CompletedResult
from astra_csp_cryptopro.settings import GlobalSettings
from astra_csp_cryptopro.settings import UserSettings
from astra_csp_cryptopro.unpack_archive_thread import UnpackArchiveThread
from astra_csp_cryptopro.utils import Utils
from astra_csp_cryptopro.verify_file_process import VerifyFileProcess
from astra_csp_cryptopro.wait_dialog import WaitDialog
from astra_csp_cryptopro.wait_thread_dialog import WaitThreadDialog

translate = QApplication.translate

mainwindow_ui = Utils.file_path('ui', 'mainwindow.ui')
UiFormClass, UiFormBase = uic.loadUiType(mainwindow_ui)  # type: QMainWindow, QMainWindow


class ResultStatus(Enum):
    Normal = 0
    Warning = 1
    Error = 2


class MainWindow(UiFormClass, UiFormBase):
    selected_document_changed = pyqtSignal(str)

    def __init__(self) -> None:
        super(MainWindow, self).__init__()
        self.tabWidget = None  # type: Optional[QTabWidget]
        self.labelDocumentPath = None  # type: Optional[QLabel]
        self.pushButtonSelectContainer = None  # type: Optional[QPushButton]
        self.pushButtonSelectDocument = None  # type: Optional[QPushButton]
        self.pushButtonDocumentView = None  # type: Optional[QPushButton]
        self.stackedWidgetSignatures = None  # type: Optional[QStackedWidget]
        self.labelSignaturePath = None  # type: Optional[QLabel]
        self.comboBoxSignatures = None  # type: Optional[QComboBox]
        self.pushButtonVerifySignature = None  # type: Optional[QPushButton]
        self.pushButtonSelectSignature = None  # type: Optional[QPushButton]
        self.textBrowserResult = None  # type: Optional[QTextBrowser]
        self.labelDocumentPathSign = None  # type: Optional[QLabel]
        self.pushButtonSelectContainerSign = None  # type: Optional[QPushButton]
        self.pushButtonSelectDocumentSign = None  # type: Optional[QPushButton]
        self.pushButtonDocumentViewSign = None  # type: Optional[QPushButton]
        self.comboBoxCertificates = None  # type: Optional[QComboBox]
        self.checkBoxTimestampCheck = None  # type: Optional[QCheckBox]
        self.checkBoxIncludeTimestamp = None  # type: Optional[QCheckBox]
        self.checkBoxCreateContainer = None  # type: Optional[QCheckBox]
        self.lineEditSignatureFileName = None  # type: Optional[QLineEdit]
        self.pushButtonSign = None  # type: Optional[QPushButton]
        self.pushButtonGenerateName = None  # type: Optional[QPushButton]
        self.textBrowserResultSign = None  # type: Optional[QTextBrowser]
        self.comboBoxFormat = None  # type: Optional[QComboBox]
        self.comboBoxExtension = None  # type: Optional[QComboBox]
        self.stackedWidgetDocument = None  # type: Optional[QStackedWidget]
        self.stackedWidgetDocumentSign = None  # type: Optional[QStackedWidget]
        self.labelContainerPath = None  # type: Optional[QLabel]
        self.labelContainerDocumentPath = None  # type: Optional[QLabel]
        self.labelContainerPathSign = None  # type: Optional[QLabel]
        self.labelContainerDocumentPathSign = None  # type: Optional[QLabel]
        self.setupUi(self)

        self.pushButtonDocumentView.clicked.connect(self.push_button_document_view_clicked)
        self.pushButtonDocumentViewSign.clicked.connect(self.push_button_document_view_sign_clicked)
        self.pushButtonSelectContainer.clicked.connect(self.push_button_select_container_clicked)
        self.pushButtonSelectDocument.clicked.connect(self.push_button_select_document_clicked)
        self.pushButtonSelectContainerSign.clicked.connect(self.push_button_select_container_sign_clicked)
        self.pushButtonSelectDocumentSign.clicked.connect(self.push_button_select_document_sign_clicked)
        self.pushButtonSelectSignature.clicked.connect(self.push_button_select_signature_clicked)
        self.pushButtonVerifySignature.clicked.connect(self.push_button_verify_signature_clicked)
        self.pushButtonSign.clicked.connect(self.push_button_sign_clicked)
        self.pushButtonGenerateName.clicked.connect(self.update_signature_filename)
        self.comboBoxExtension.currentIndexChanged.connect(self.update_signature_extension)

        self.comboBoxSignatures.activated.connect(self.on_signatures_combobox_index_changed)
        self.comboBoxCertificates.currentIndexChanged.connect(self.clear_sign_result)
        self.comboBoxFormat.currentIndexChanged.connect(self.clear_sign_result)
        self.comboBoxExtension.currentIndexChanged.connect(self.clear_sign_result)
        self.lineEditSignatureFileName.textChanged.connect(self.clear_sign_result)

        self.success_palette = QGuiApplication.palette()
        self.success_palette.setColor(QPalette.Base, QColor("#DDFFDE"))
        self.success_palette.setColor(QPalette.Text, QColor("#000000"))
        self.warning_palette = QGuiApplication.palette()
        self.warning_palette.setColor(QPalette.Base, QColor("yellow"))
        self.warning_palette.setColor(QPalette.Text, QColor("#000000"))
        self.error_palette = QGuiApplication.palette()
        self.error_palette.setColor(QPalette.Base, QColor("#FE4543"))
        self.error_palette.setColor(QPalette.Text, QColor("#FFFFFF"))

        icon = QIcon.fromTheme('fly-csp-cryptopro')
        self.setWindowIcon(icon)
        self.tabWidget.setCurrentIndex(0)
        self.stackedWidgetDocument.setCurrentIndex(0)
        self.stackedWidgetDocumentSign.setCurrentIndex(0)
        self.stackedWidgetSignatures.setCurrentIndex(0)

        self.container = None  # type: Optional[Container]
        self.document = None  # type: Optional[str]
        self.signature = None  # type: Optional[str]
        self.user_settings = None  # type: Optional[UserSettings]
        self.global_settings = None  # type: Optional[GlobalSettings]
        self.unpack_container_thread = None  # type: Optional[UnpackArchiveThread]
        self.enable_extension_filter = False  # type: bool
        self.extensions_enabled_for_signing = []  # type: List[str]
        self.sigextensions = ['.p7s', '.sig', '.sign']
        self.lastdir = QDir.homePath()
        self.cache = ApplicationCache()
        self.cryptopro_bin_dir = None  # type: Optional[QDir]

        self.comboBoxCertificates.setCurrentText(translate('text', 'Loading...'))

        self.waiting_dialog = None  # type: Optional[WaitDialog]
        self.waiting_timer = QTimer(self)
        self.waiting_timer.setSingleShot(True)
        self.waiting_timer.setInterval(2000)
        self.waiting_timer.timeout.connect(self.show_waiting_dialog)

        self.cached_time_mark_plugin_installed_value = None  # type: Optional[bool]

    # noinspection PyPep8Naming
    def closeEvent(self, event: QCloseEvent) -> None:
        self.stop_unpack_container_files()
        self.save_selected_format()
        self.save_selected_extension()
        self.save_last_dir()
        self.clear_cache()
        event.accept()
        # Workaround app running forever after close event in document selection mode
        QApplication.quit()

    # noinspection PyPep8Naming
    def keyPressEvent(self, keyevent: QKeyEvent) -> None:
        if keyevent.key() == Qt.Key_F1:
            self.show_help()

    def initialize(self, files: List[str]) -> None:
        try:
            self.unify_buttons_sizes()
            self.user_settings = UserSettings()
            self.global_settings = GlobalSettings()
            self.setup_cryptopro_bin_dir()
            self.validate_system()
            self.enable_extension_filter = self.global_settings.enable_extension_filter
            self.extensions_enabled_for_signing = self.global_settings.extensions_enabled_for_signing
            self.setup_formats()
            self.setup_extensions()
            self.setup_last_dir()
            self.cache.create_dir()

            plugin_loader = PluginLoader()
            plugin_loader.load()
            for tab_plugin in plugin_loader.tab_plugins():
                self.tabWidget.addTab(tab_plugin.create_widget(self), tab_plugin.label())

            self.cached_time_mark_plugin_installed_value = self.time_mark_plugin_installed()
            if self.cached_time_mark_plugin_installed_value:
                self.checkBoxTimestampCheck.setToolTip(
                    translate('text', 'If the signature was created with a time stamp, '
                                      'then you need to activate this checkbox'))
                self.checkBoxIncludeTimestamp.stateChanged.connect(self.on_check_box_include_timestamp_state_changed)
            else:
                self.checkBoxTimestampCheck.setEnabled(False)
                self.checkBoxTimestampCheck.setChecked(False)
                timestamp_tooltip_text = translate('text',
                                                   'Package cprocsp-pki-cades or cprocsp-pki-cades-64 not installed')
                self.checkBoxTimestampCheck.setToolTip(timestamp_tooltip_text)

                self.checkBoxIncludeTimestamp.setEnabled(False)
                self.checkBoxIncludeTimestamp.setChecked(False)
                self.checkBoxIncludeTimestamp.setToolTip(timestamp_tooltip_text)

            self.perform_input_files(files)
            self.toggle_signature_tab()

            merge_certificates_process = MergeCertificatesProcess(bin_dir=self.cryptopro_bin_dir)
            merge_certificates_process.kill_time = 0
            merge_certificates_process.completed.connect(self.on_merge_certificates_process_completed)
            merge_wait_dialog = CprocspProcessDialog(process=merge_certificates_process, parent=self)
            self.start_waiting(merge_wait_dialog)
            merge_wait_dialog.setWindowTitle(translate('text', 'Adding user certificates to certificate store'))
            merge_certificates_process.start()
        except Exception as exception:
            self.show_critical_error(translate('text', 'Critical error'),
                                     translate('text', 'Failed to initialize system'),
                                     str(exception))

    def on_application_message_received(self, message: str) -> None:
        try:
            self.application_message_received(message)
        except Exception as exception:
            self.show_error(translate('text', 'Failed perform application message'),
                            str(exception))

    def application_message_received(self, message: str) -> None:
        if self.waiting_dialog:
            QMessageBox.warning(self,
                                translate('text', 'Operation in progress'),
                                translate('text', 'Wait for current operation to complete'))
            return

        files = message.split('\t')
        self.perform_input_files(files)
        self.maybe_verify()

    def perform_input_files(self, files: List[str]) -> None:
        if not files:
            return

        self.clear_verification_result()

        containers, documents, signatures = self.get_input_files(files)
        if containers:
            self.change_selected_container(containers[0])
        else:
            if documents:
                self.change_selected_document(documents[0])
            if signatures:
                if not documents:
                    self.change_selected_document(signatures[0])
                else:
                    self.change_signature_file(signatures[0])

    def start_unpack_container_files(self) -> None:
        if not self.container:
            return

        self.unpack_container_thread = self.container.create_unpack_thread()
        self.unpack_container_thread.finished.connect(self.on_unpack_container_thread_finished)
        self.unpack_container_thread.start()

    def stop_unpack_container_files(self) -> None:
        if self.unpack_container_thread:
            if self.unpack_container_thread.isRunning():
                self.unpack_container_thread.terminate()
            self.unpack_container_thread = None

    def on_unpack_container_thread_finished(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        if thread.error is not None:
            self.show_error(translate('text', 'Container error'),
                            translate('text', 'Failed to unpack container'),
                            thread.error,
                            stop_waiting=False)

    def on_merge_certificates_process_completed(self, result: CompletedResult) -> None:
        process = self.sender()
        if type(process) != MergeCertificatesProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            self.emergency_exit()
        elif result.error is not None:
            self.show_error(translate('text', 'Process error'),
                            translate('text', 'Failed to add user certificates to certificate store'),
                            process.errorString())
        elif result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.show_error(translate('text', 'Process crashed'),
                            translate('text', 'Failed to add user certificates to certificate store'),
                            process.errorString())
        elif result.exit_code is not None and result.exit_code != 0:
            self.show_error(translate('text', 'Process exited with unexpected code: {code}')
                            .format(code=result.exit_code),
                            translate('text', 'Failed to add user certificates to certificate store'),
                            process.output())

        self.enable_ui()
        certificates_process = CertificateProcess(bin_dir=self.cryptopro_bin_dir)
        certificates_process.kill_time = self.global_settings.max_csp_time_waiting
        certificates_process.completed.connect(self.on_get_certificates_process_completed)
        get_certificates_dialog = CprocspProcessDialog(process=certificates_process, parent=self)
        self.start_waiting(get_certificates_dialog)
        get_certificates_dialog.setWindowTitle(translate('text', 'Loading certificates'))
        certificates_process.start()

    def on_get_certificates_process_completed(self, result: CompletedResult) -> None:
        process = self.sender()
        if type(process) != CertificateProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            self.emergency_exit()
        elif result.error is not None:
            self.show_critical_error(translate('text', 'Process error'),
                                     translate('text', 'Failed to load certificates'),
                                     process.errorString())
            return
        elif result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.show_critical_error(translate('text', 'Process crashed'),
                                     translate('text', 'Failed to load certificates'),
                                     process.errorString())
            return
        elif process.empty_list_error_occurred():
            self.show_critical_error(translate('text', 'Failed to load certificates'),
                                     translate('text', 'Certificates not available. Empty certificate list.'))
            return
        elif result.exit_code is not None and result.exit_code != 0:
            self.show_critical_error(translate('text', 'Failed to load certificates'),
                                     translate('text', 'Process exited with unexpected code: {code}')
                                     .format(code=result.exit_code),
                                     process.output())
            return
        else:
            self.setup_certificates(process.get_certificates())

        self.enable_ui()
        self.maybe_verify()

    def on_create_file_signature_process_completed(self, result: CompletedResult) -> None:
        process = self.sender()
        if type(process) != CreateFileSignatureProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            self.clean_cancelled_signature(process.signature)
            return
        if result.error is not None:
            self.show_error(translate('text', 'Failed to create file signature'),
                            process.errorString(),
                            process.errorString())
            self.clean_cancelled_signature(process.signature)
            return
        if result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.show_error(translate('text', 'Process crashed'),
                            translate('text', 'Failed to create file signature'),
                            process.errorString())
            self.clean_cancelled_signature(process.signature)
            return

        if process.timestamp_error_occurred():
            status = ResultStatus.Warning
        elif process.failed_output_error_code_occurred():
            status = ResultStatus.Error
        else:
            status = ResultStatus.Normal

        self.show_signing_result(process.output(), status)

        if status == ResultStatus.Error:
            return
        elif self.checkBoxCreateContainer.isChecked():
            self.create_container(process.document,
                                  process.signature)
        elif self.container:
            container_creation_thread = ContainerCreationThread(container=self.container,
                                                                files=[process.document, process.signature],
                                                                update=True)
            container_creation_thread.finished.connect(self.on_update_container_thread_finished)
            wait_dialog = WaitThreadDialog(thread=container_creation_thread, parent=self)
            self.start_waiting(wait_dialog)
            container_creation_thread.start()
        else:
            self.change_signature_file(process.signature)

    def on_update_container_thread_finished(self):
        thread = self.sender()
        if type(thread) != ContainerCreationThread:
            return

        self.stop_waiting()
        if thread.error is not None:
            self.show_error(translate('text', 'Container error'),
                            translate('text', 'Failed to update container'),
                            thread.error)
        elif not thread.terminated:
            self.change_signature_file(self.container.mount_dir.path() + '/' + self.lineEditSignatureFileName.text())

    def clean_cancelled_signature(self, signature: str) -> None:
        try:
            if Utils.file_is_exists(signature):
                os.remove(signature)
        except Exception as exception:
            self.show_error(translate('text', 'Signing cancel error'),
                            translate('text', 'Failed to remove signature file'),
                            str(exception))

    def create_container(self, document: str, signature: str) -> None:
        name = document + '.signed.zip'
        container = Container(name, self.cache.common_cache_path())

        thread = ContainerCreationThread(container, [document, signature])
        thread.finished.connect(self.on_container_creation_thread_finished)
        wait_dialog = WaitThreadDialog(thread=thread, parent=self)
        self.start_waiting(wait_dialog)

        thread.start()

    def on_container_creation_thread_finished(self) -> None:
        thread = self.sender()
        if type(thread) != ContainerCreationThread:
            return

        self.stop_waiting()
        if thread.error is not None:
            self.show_error(translate('text', 'Container error'),
                            translate('text', 'Failed to create container'),
                            thread.error)
        elif not thread.terminated:
            self.lineEditSignatureFileName.blockSignals(True)

            try:
                os.remove(self.document)
                os.remove(os.path.dirname(thread.container.path) + '/' + self.lineEditSignatureFileName.text())
            except Exception as exception:
                self.show_error(translate('text', 'Container error'),
                                translate('text', 'Failed to remove original files'),
                                str(exception))

            document_basename = os.path.basename(self.document)
            signature_base_name = self.lineEditSignatureFileName.text()
            self.change_selected_container(thread.container.path, False)
            self.change_selected_document(self.container.mount_dir.path() + '/' + document_basename, False)
            self.change_signature_file(self.container.mount_dir.path() + '/' + signature_base_name)
            self.lineEditSignatureFileName.blockSignals(False)

    def on_pdf_conversion_process_completed(self, result: CompletedResult) -> None:
        process = self.sender()
        if type(process) != PdfConversionProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            return
        if result.error is not None:
            self.show_error(translate('text', 'Process error'),
                            translate('text', 'Failed to open file'),
                            process.errorString())
            return
        if result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.show_error(translate('text', 'Process crashed'),
                            translate('text', 'Failed to open file'),
                            process.errorString())
            return
        if result.exit_code is not None and result.exit_code != 0:
            self.show_error(translate('text', 'Failed to open file'),
                            translate('text', 'Process exited with unexpected code: {code}')
                            .format(code=result.exit_code),
                            Utils.to_utf8(process.readAll()))
            return

        Utils.open_file(process.output_file_path(file_path=self.document,
                                                 outdir=self.cache.cache_path()))

    def on_verify_file_process_completed(self, result: CompletedResult) -> None:
        process = self.sender()
        if type(process) != VerifyFileProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            return
        if result.error is not None:
            self.show_error(translate('text', 'Process error'),
                            translate('text', 'Failed to verify file'),
                            process.errorString())
            return
        if result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.show_error(translate('text', 'Process crashed'),
                            translate('text', 'Failed to verify file'),
                            process.errorString())
            return

        if process.timestamp_error_occurred():
            status = ResultStatus.Warning
        elif process.failed_output_error_code_occurred():
            status = ResultStatus.Error
        else:
            status = ResultStatus.Normal

        self.show_verification_result(process.output(), status)

    def setup_certificates(self, certificates: List[Certificate]) -> None:
        self.comboBoxCertificates.clear()
        fprint = self.user_settings.fingerprint
        for certificate in sorted(certificates, key=lambda cert: str(cert.subject)):
            subject = str(certificate.subject)
            self.comboBoxCertificates.addItem(subject, certificate)
            if fprint and certificate.fingerprint == fprint:
                self.comboBoxCertificates.setCurrentText(subject)

    def setup_formats(self) -> None:
        selected_format = self.user_settings.selected_format
        if isinstance(selected_format, str):
            self.comboBoxFormat.setCurrentText(selected_format)

    def setup_extensions(self) -> None:
        extensions = self.user_settings.extensions or self.sigextensions
        for extension in extensions:
            self.comboBoxExtension.addItem(extension)
        selected_extension = self.user_settings.selected_extension
        if isinstance(selected_extension, str):
            self.comboBoxExtension.setCurrentText(selected_extension)

    def current_extensions(self) -> List[str]:
        extensions = []
        for i in range(0, self.comboBoxExtension.count()):
            extensions.append(self.comboBoxExtension.itemText(i))
        return extensions

    def maybe_add_extenstion(self, extension: str) -> None:
        if not extension:
            return

        extensions = set(self.current_extensions())
        if extension not in extensions:
            self.comboBoxExtension.addItem(extension)
        self.comboBoxExtension.setCurrentText(extension)

    def save_extensions(self) -> None:
        self.user_settings.extensions = self.current_extensions()

    def save_selected_format(self) -> None:
        self.user_settings.selected_format = self.comboBoxFormat.currentText()

    def save_selected_extension(self) -> None:
        self.user_settings.selected_extension = self.comboBoxExtension.currentText()

    def setup_last_dir(self) -> None:
        lastdir = self.user_settings.last_dir
        if isinstance(lastdir, str):
            self.lastdir = lastdir or QDir.homePath()

    def document_last_dir(self) -> str:
        if self.container and Utils.dir_is_exists(os.path.dirname(self.container.path)):
            return os.path.dirname(self.container.path)
        elif self.document and Utils.dir_is_exists(os.path.dirname(self.document)):
            return os.path.dirname(self.document)
        else:
            return self.lastdir

    def signature_last_dir(self) -> str:
        if self.document and Utils.dir_is_exists(os.path.dirname(self.document)):
            return os.path.dirname(self.document)
        else:
            return self.lastdir

    def save_last_dir(self) -> None:
        if self.container and Utils.dir_is_exists(os.path.dirname(self.container.path)):
            self.lastdir = os.path.dirname(self.container.path)
        elif self.document and Utils.dir_is_exists(os.path.dirname(self.document)):
            self.lastdir = os.path.dirname(self.document)
        self.user_settings.last_dir = self.lastdir

    # Разделяет входящие файлы на контейнеры, документы и подписи
    def get_input_files(self, files: List[str]) -> Tuple[List[str], List[str], List[str]]:
        containers = []
        documents = []
        signatures = []
        for file in files:
            extension = os.path.splitext(file)[1]
            if extension in self.current_extensions():
                signatures.append(file)
            elif Container.is_container(file):
                containers.append(file)
            elif Container.is_external_container(file):
                containers.append(file)
            else:
                documents.append(file)
        return containers, documents, signatures

    def maybe_verify(self) -> None:
        if self.container:
            return

        if bool(self.document) and Utils.file_is_exists(self.document) \
                and bool(self.signature) and Utils.file_is_exists(self.signature):
            self.verify_document()

    def is_valid(self) -> bool:
        if not self.document:
            return False

        if not self.comboBoxCertificates.count() > 0:
            return False

        if self.container:
            if not self.container.contains(self.document):
                return False

        return True

    def enable_signature_actions(self) -> None:
        valid = self.is_valid()
        self.pushButtonDocumentView.setEnabled(valid)
        self.pushButtonDocumentViewSign.setEnabled(valid)
        self.lineEditSignatureFileName.setEnabled(valid)
        self.pushButtonGenerateName.setEnabled(valid)
        self.pushButtonSign.setEnabled(valid)
        self.pushButtonVerifySignature.setEnabled(valid)
        self.pushButtonSelectSignature.setEnabled(valid)
        self.checkBoxTimestampCheck.setEnabled(bool(self.cached_time_mark_plugin_installed_value))

    def enable_ui(self) -> None:
        self.tabWidget.setEnabled(True)
        self.enable_signature_actions()

    def enable_controls(self) -> None:
        self.pushButtonSelectContainer.setEnabled(True)
        self.pushButtonSelectContainerSign.setEnabled(True)
        self.pushButtonSelectDocument.setEnabled(True)
        self.pushButtonSelectDocumentSign.setEnabled(True)
        self.pushButtonDocumentView.setEnabled(True)
        self.pushButtonDocumentViewSign.setEnabled(True)
        self.pushButtonSelectSignature.setEnabled(True)
        self.comboBoxCertificates.setEnabled(True)
        self.comboBoxFormat.setEnabled(True)
        self.comboBoxExtension.setEnabled(True)
        if self.cached_time_mark_plugin_installed_value:
            self.checkBoxIncludeTimestamp.setEnabled(True)
        self.enable_signature_actions()
        if not self.container:
            self.checkBoxCreateContainer.setEnabled(True)
        else:
            self.checkBoxCreateContainer.setEnabled(False)

    def disable_controls(self) -> None:
        self.pushButtonSign.setDisabled(True)
        self.pushButtonSelectContainer.setDisabled(True)
        self.pushButtonSelectContainerSign.setDisabled(True)
        self.pushButtonSelectDocument.setDisabled(True)
        self.pushButtonSelectDocumentSign.setDisabled(True)
        self.pushButtonDocumentView.setDisabled(True)
        self.pushButtonDocumentViewSign.setDisabled(True)
        self.pushButtonGenerateName.setDisabled(True)
        self.pushButtonSelectSignature.setDisabled(True)
        self.pushButtonVerifySignature.setDisabled(True)
        self.comboBoxCertificates.setDisabled(True)
        self.comboBoxFormat.setDisabled(True)
        self.comboBoxExtension.setDisabled(True)
        self.checkBoxTimestampCheck.setDisabled(True)
        self.checkBoxIncludeTimestamp.setDisabled(True)
        self.lineEditSignatureFileName.setDisabled(True)
        self.checkBoxCreateContainer.setDisabled(True)

    def toggle_signature_tab(self) -> None:
        if self.enable_extension_filter:
            if not self.document:
                self.tabWidget.setCurrentIndex(0)
                self.tabWidget.setTabEnabled(1, False)
            elif os.path.splitext(self.document)[1] not in self.extensions_enabled_for_signing:
                self.tabWidget.setCurrentIndex(0)
                self.tabWidget.setTabEnabled(1, False)
            else:
                self.tabWidget.setTabEnabled(1, True)

    def start_wait_container_unpacking(self) -> bool:
        if not self.unpack_container_thread:
            self.start_unpack_container_files()
            if not self.waiting_dialog:
                self.start_waiting(WaitThreadDialog(thread=self.unpack_container_thread, parent=self))
            return True

        if self.unpack_container_thread.isRunning():
            if not self.waiting_dialog:
                self.start_waiting(WaitThreadDialog(thread=self.unpack_container_thread, parent=self))
            return True
        else:
            self.stop_unpack_container_files()

        return False

    def push_button_document_view_clicked(self, _: bool) -> None:
        self.perform_document_view_clicked()

    def push_button_document_view_sign_clicked(self, _: bool) -> None:
        self.perform_document_view_clicked()

    def unpack_container_thread_finished_after_view_clicked(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        self.stop_waiting()
        self.stop_unpack_container_files()

        if not thread.terminated and not thread.error:
            self.view_document()

    def is_container_exists(self) -> bool:
        if self.container:
            return Utils.file_is_exists(self.container.path)
        else:
            return True

    def check_container_existance(self) -> None:
        if not self.is_container_exists():
            raise Exception(translate('text', 'Container is not available'))

    def perform_document_view_clicked(self) -> None:
        try:
            self.check_container_existance()
            if self.container \
                    and not self.container.unpacked() \
                    and self.start_wait_container_unpacking():
                self.unpack_container_thread.finished.connect(self.unpack_container_thread_finished_after_view_clicked)
            else:
                self.view_document()
        except Exception as exception:
            self.show_error(translate('text', 'Failed to open document'),
                            str(exception))

    def view_document(self) -> None:
        self.save_last_dir()
        if Utils.file_mimetypes(self.document) in self.global_settings.pdf_conversion_mimetypes:
            cache_document_path = PdfConversionProcess.output_file_path(file_path=self.document,
                                                                        outdir=self.cache.cache_path())
            if Utils.file_is_exists(cache_document_path):
                Utils.open_file(cache_document_path)
            else:
                sizelimit = 1024 * 1024 * 10  # 10 MB
                if os.path.getsize(self.document) < sizelimit or self.ask_about_conversion():
                    pdf_conversion_process = PdfConversionProcess()
                    pdf_conversion_process.kill_time = 0  # Let the user decide if he wants to convert large files
                    pdf_conversion_process.completed.connect(self.on_pdf_conversion_process_completed)
                    pdf_conversion_process.setup_arguments(file_path=self.document,
                                                           outdir=self.cache.cache_path())
                    self.start_waiting(InteractiveProcessDialog(process=pdf_conversion_process, parent=self))
                    pdf_conversion_process.start()
        else:
            Utils.open_file(self.document)

    def push_button_verify_signature_clicked(self, _: bool) -> None:
        try:
            self.check_container_existance()
            if self.container \
                    and not self.container.unpacked() \
                    and self.start_wait_container_unpacking():
                self.unpack_container_thread.finished.connect(
                    self.unpack_container_thread_finished_after_verify_clicked)
            else:
                self.verify_document()
        except Exception as exception:
            self.show_error(translate('text', 'Failed to verify document'),
                            str(exception))

    def unpack_container_thread_finished_after_verify_clicked(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        self.stop_waiting()
        self.stop_unpack_container_files()

        if not thread.terminated and not thread.error:
            self.verify_document()

    def verify_document(self) -> None:
        self.save_last_dir()
        self.clear_verification_result()
        verify_file_process = VerifyFileProcess(bin_dir=self.cryptopro_bin_dir)
        verify_file_process.kill_time = 0  # Signature verification time depends on file size
        verify_file_process.completed.connect(self.on_verify_file_process_completed)
        verify_file_process.setup_arguments(file_path=self.document,
                                            check_timestamp=self.check_timestamp(),
                                            sig_path=self.signature)
        self.start_waiting(CprocspProcessDialog(process=verify_file_process, parent=self))
        verify_file_process.start()

    def push_button_sign_clicked(self, _: bool) -> None:
        try:
            self.check_container_existance()
            if self.container \
                    and not self.container.unpacked() \
                    and self.start_wait_container_unpacking():
                self.unpack_container_thread.finished.connect(self.unpack_container_thread_finished_after_sign_clicked)
            else:
                self.sign_document()
        except Exception as exception:
            self.show_error(translate('text', 'Failed to sign document'),
                            str(exception))

    def unpack_container_thread_finished_after_sign_clicked(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        self.stop_waiting()
        self.stop_unpack_container_files()

        if not thread.terminated and not thread.error:
            self.sign_document()

    def sign_document(self) -> None:
        self.clear_sign_result()
        dirname = os.path.dirname(self.document)
        filename = self.lineEditSignatureFileName.text()
        extension = os.path.splitext(filename)[1]
        filepath = "{dir}/{file}".format(dir=dirname, file=filename)
        fprint = self.comboBoxCertificates.currentData().fingerprint
        self.save_fingerprint(fprint)
        self.maybe_add_extenstion(extension)
        self.save_extensions()
        self.save_selected_extension()
        self.save_selected_format()
        self.save_last_dir()
        create_file_signature_process = CreateFileSignatureProcess(bin_dir=self.cryptopro_bin_dir)
        create_file_signature_process.kill_time = 0  # Signing time depends on file size
        create_file_signature_process.completed.connect(self.on_create_file_signature_process_completed)
        create_file_signature_process.setup_arguments(file_path=self.document,
                                                      sig_name=filepath,
                                                      fingerprint=fprint,
                                                      der_format=self.comboBoxFormat.currentText() == 'DER',
                                                      time_address=self.global_settings.time_address
                                                      if self.include_timestamp() else None)
        self.start_waiting(CprocspProcessDialog(process=create_file_signature_process, parent=self))
        create_file_signature_process.start()

    def save_fingerprint(self, fprint: str) -> None:
        self.user_settings.fingerprint = fprint

    def update_signature_filename(self) -> None:
        if self.document:
            name = self.generate_signature_name(self.document)
            self.lineEditSignatureFileName.setText(name)

    def update_signature_extension(self) -> None:
        path = self.lineEditSignatureFileName.text()
        if path:
            root = os.path.splitext(path)[0]
            ext = self.comboBoxExtension.currentText()
            self.lineEditSignatureFileName.setText(root + ext)

    def generate_signature_name(self, path: str) -> str:
        date = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        base = os.path.basename(path)
        ext = self.comboBoxExtension.currentText()
        if self.include_timestamp():
            return "{base}_{date}.tsp{ext}".format(base=base, date=date, ext=ext)
        else:
            return "{base}_{date}{ext}".format(base=base, date=date, ext=ext)

    def include_timestamp(self) -> bool:
        return self.checkBoxIncludeTimestamp.isChecked()

    def check_timestamp(self) -> bool:
        return self.checkBoxTimestampCheck.isChecked()

    def push_button_select_container_clicked(self, _: bool) -> None:
        self.select_container_clicked_safe()

    def push_button_select_container_sign_clicked(self, _: bool) -> None:
        self.select_container_clicked_safe()

    def select_container_clicked_safe(self) -> None:
        try:
            self.select_container_clicked()
        except Exception as exception:
            self.show_error(translate('text', 'Failed to select container'),
                            str(exception))

    def select_container_clicked(self) -> None:
        startdir = self.document_last_dir()
        dialog = QFileDialog(self, translate('text', 'Select container'), startdir)
        extensions = ['*.zip']
        formatted_extensions = ' '.join(extensions)
        dialog.setNameFilters([translate('text', 'Containers ({exts})').format(exts=formatted_extensions)])
        if dialog.exec():
            self.change_selected_container(dialog.selectedFiles()[0])

    def change_selected_container(self, file_path: Optional[str], clear: bool = True) -> None:
        self.stop_unpack_container_files()

        self.container = None
        if file_path:
            try:
                if Container.is_external_container(file_path):
                    archive, mountpoint = Container.find_external(file_path)
                    Container.check_encryption(archive)
                    Container.unmount(archive, mountpoint)
                    self.container = Container(filepath=archive, tmpdir=self.cache.common_cache_path())
                else:
                    Container.check_encryption(file_path)
                    Container.find_and_unmount(file_path)
                    self.container = Container(filepath=file_path, tmpdir=self.cache.common_cache_path())
            except UnmountError as exception:
                self.show_error(translate('text', 'Unable to open the container'),
                                translate('text', 'Unable to open the container, since it is already in use'),
                                str(exception.details))
            except ContainerException as exception:
                self.show_error(translate('text', 'Unable to open the container'),
                                exception.message,
                                str(exception.details))
            except Exception as exception:
                self.show_error(translate('text', 'Unable to open the container'),
                                translate('text', 'Unable to open the container'),
                                str(exception))

        if self.container:
            self.labelContainerPath.setText(self.container.path)
            self.labelContainerPathSign.setText(self.container.path)
            self.stackedWidgetDocument.setCurrentIndex(1)
            self.stackedWidgetDocumentSign.setCurrentIndex(1)
            self.stackedWidgetSignatures.setCurrentIndex(1)
            self.checkBoxCreateContainer.setChecked(False)
            self.checkBoxCreateContainer.setEnabled(False)
        else:
            self.labelContainerPath.clear()
            self.labelContainerPathSign.clear()
            self.stackedWidgetDocument.setCurrentIndex(0)
            self.stackedWidgetDocumentSign.setCurrentIndex(0)
            self.stackedWidgetSignatures.setCurrentIndex(0)
            self.checkBoxCreateContainer.setChecked(False)
            self.checkBoxCreateContainer.setEnabled(True)

        if self.container:
            container_basename = QFileInfo(self.container.name).baseName()
            signature_extensions = [extension[1:] for extension in self.sigextensions]
            document = None
            signatures = []
            for filename in self.container.files():
                file_info = QFileInfo(filename)
                if file_info.suffix() in signature_extensions:
                    signatures.append(self.container.sources_path() + '/' + filename)
                elif file_info.baseName() == container_basename:
                    document = self.container.sources_path() + '/' + filename
            self.setup_signatures(signatures)
            self.change_selected_document(document, clear)

        if self.container:
            self.start_unpack_container_files()

    def setup_signatures(self, signatures: List[str]) -> None:
        self.comboBoxSignatures.clear()
        self.comboBoxSignatures.addItem(translate('text', 'Attached digital signature'), None)
        for index, signature in enumerate(signatures):
            signature_name = os.path.basename(signature)
            self.comboBoxSignatures.addItem(signature_name, signature)
            self.comboBoxSignatures.setItemData(index + 1, signature, Qt.ToolTipRole)
        if self.comboBoxSignatures.count() > 1:
            self.comboBoxSignatures.setCurrentIndex(1)

    def push_button_select_document_clicked(self, _: bool) -> None:
        self.select_document_clicked_safe()

    def push_button_select_document_sign_clicked(self, _: bool) -> None:
        self.select_document_clicked_safe()

    def select_document_clicked_safe(self) -> None:
        try:
            self.select_document_clicked()
        except Exception as exception:
            self.show_error(translate('text', 'Failed select document'),
                            str(exception))

    def select_document_clicked(self) -> None:
        if not self.is_container_exists():
            self.show_error(translate('text', 'Unable to select file in the container'),
                            translate('text', 'Container is not available'))

        if self.container \
                and Utils.file_is_exists(self.container.path) \
                and not self.container.unpacked() \
                and self.start_wait_container_unpacking():
            self.unpack_container_thread.finished.connect(
                self.unpack_container_thread_finished_after_select_document_clicked)
            return

        self.show_select_document_dialog()

    def unpack_container_thread_finished_after_select_document_clicked(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        self.stop_waiting()
        self.stop_unpack_container_files()

        if not thread.terminated and not thread.error:
            self.show_select_document_dialog()

    def start_dir(self) -> str:
        if self.container:
            if Utils.file_is_exists(self.container.path):
                return self.container.mount_dir.path()
            else:
                return self.document_last_dir()
        else:
            return self.document_last_dir()

    def show_select_document_dialog(self) -> None:
        dialog = QFileDialog(self, translate('text', 'Select document'), self.start_dir())
        if dialog.exec():
            self.change_selected_document(dialog.selectedFiles()[0])

    def change_selected_document(self, file_path: Optional[str], clear: bool = True) -> None:
        self.document = file_path

        if self.container and self.document and not self.container.contains(self.document):
            self.change_selected_container(None, clear)

        self.pushButtonDocumentView.setEnabled(bool(self.document))
        self.pushButtonDocumentViewSign.setEnabled(bool(self.document))
        self.pushButtonGenerateName.setEnabled(bool(self.document))
        self.lineEditSignatureFileName.setEnabled(bool(self.document))
        self.pushButtonSign.setEnabled(bool(self.document))
        self.checkBoxCreateContainer.setChecked(False)
        self.checkBoxCreateContainer.setEnabled(bool(self.document) and not bool(self.container))

        if self.document:
            self.lineEditSignatureFileName.setText(self.generate_signature_name(self.document))
            if self.container:
                self.labelContainerDocumentPath.setText(os.path.basename(self.document))
                self.labelContainerDocumentPathSign.setText(os.path.basename(self.document))
                self.labelDocumentPath.clear()
                self.labelDocumentPathSign.clear()
                self.stackedWidgetDocument.setCurrentIndex(1)
                self.stackedWidgetDocumentSign.setCurrentIndex(1)
                self.stackedWidgetSignatures.setCurrentIndex(1)
            else:
                self.labelDocumentPath.setText(self.document)
                self.labelDocumentPathSign.setText(self.document)
                self.labelContainerDocumentPath.clear()
                self.labelContainerDocumentPathSign.clear()
                self.stackedWidgetDocument.setCurrentIndex(0)
                self.stackedWidgetDocumentSign.setCurrentIndex(0)
                self.stackedWidgetSignatures.setCurrentIndex(0)
        else:
            self.lineEditSignatureFileName.clear()
            self.labelDocumentPath.clear()
            self.labelDocumentPathSign.clear()
            self.labelContainerDocumentPath.clear()
            self.labelContainerDocumentPathSign.clear()

        if self.container:
            self.on_signatures_combobox_index_changed()
        else:
            self.change_signature_file(None)
        self.toggle_signature_tab()

        if clear:
            self.clear_sign_result()
        self.clear_verification_result()

    def update_signature_text(self) -> None:
        if self.container and self.signature and os.path.dirname(self.signature) == self.container.sources_path():
            self.labelSignaturePath.setText(os.path.basename(self.signature))
        elif self.signature:
            self.labelSignaturePath.setText(self.signature)
        elif self.document:
            self.labelSignaturePath.setText(translate('text', 'Attached digital signature will be verified'))
        else:
            self.labelSignaturePath.clear()

    def push_button_select_signature_clicked(self) -> None:
        self.select_signature_clicked_safe()

    def select_signature_clicked_safe(self) -> None:
        try:
            self.select_signature_clicked()
        except Exception as exception:
            self.show_error(translate('text', 'Failed select signature'),
                            str(exception))

    def select_signature_clicked(self) -> None:
        self.check_container_existance()
        if self.container \
                and not self.container.unpacked() \
                and self.start_wait_container_unpacking():
            self.unpack_container_thread.finished.connect(
                self.unpack_container_thread_finished_after_select_signature_clicked)
            return

        self.show_select_signature_dialog()

    def on_signatures_combobox_index_changed(self) -> None:
        path = self.comboBoxSignatures.currentData()
        self.change_signature_file(path)

    def add_signature_to_combobox(self, signature: str) -> None:
        if signature is None:
            return

        found = False
        for i in range(0, self.comboBoxSignatures.count()):
            path = self.comboBoxSignatures.itemData(i)
            if path == signature:
                self.comboBoxSignatures.setCurrentIndex(i)
                found = True
                break
        if not found:
            base = os.path.basename(signature)
            self.comboBoxSignatures.addItem(base, signature)
            index = self.comboBoxSignatures.count() - 1
            self.comboBoxSignatures.setItemData(index, signature, Qt.ToolTipRole)
            self.comboBoxSignatures.setCurrentIndex(index)

    def unpack_container_thread_finished_after_select_signature_clicked(self) -> None:
        thread = self.sender()
        if type(thread) != UnpackArchiveThread:
            return

        self.stop_waiting()
        self.stop_unpack_container_files()

        if not thread.terminated and not thread.error:
            self.show_select_signature_dialog()

    def show_select_signature_dialog(self) -> None:
        dialog = QFileDialog(self, translate('text', 'Select signature'), self.start_dir())
        extenstions = []
        for extension in self.current_extensions():
            extenstions.append('*' + extension)
        formatted_extensions = ' '.join(extenstions)
        dialog.setNameFilters([translate('text', 'Signature files ({exts})').format(exts=formatted_extensions),
                               translate('text', 'All files (*)')])
        if dialog.exec():
            self.change_signature_file(dialog.selectedFiles()[0])

    def change_signature_file(self, file_path: Optional[str]) -> None:
        self.signature = file_path

        self.update_signature_text()
        self.clear_verification_result()
        self.enable_signature_actions()
        if self.signature_contains_timestamp_extension():
            if not self.cached_time_mark_plugin_installed_value:
                self.checkBoxTimestampCheck.setChecked(False)
                QMessageBox.warning(self,
                                    translate('text', 'Warning'),
                                    translate('text', 'Timestamped signature file checking not possible. '
                                                      'Package cprocsp-pki-cades not installed. '
                                                      'An file signature check will be performed without a timestamp.'))
            else:
                self.checkBoxTimestampCheck.setChecked(True)
        else:
            self.checkBoxTimestampCheck.setChecked(False)
        self.add_signature_to_combobox(file_path)

    def signature_contains_timestamp_extension(self) -> bool:
        if not self.signature:
            return False

        basename = os.path.basename(self.signature)
        parts = basename.split('.')
        if len(parts) > 1:
            return 'tsp' in parts[1:]
        else:
            return False

    def show_verification_result(self, result: str, status: ResultStatus) -> None:
        self.show_result(self.textBrowserResult, result, status)

    def show_signing_result(self, result: str, status: ResultStatus) -> None:
        self.show_result(self.textBrowserResultSign, result, status)

    def show_result(self, widget: QTextBrowser, result: str, status: ResultStatus) -> None:
        widget.setPalette(self.status_palette(status))
        widget.setText(result)

    def clear_verification_result(self) -> None:
        self.textBrowserResult.clear()
        self.textBrowserResult.setPalette(QGuiApplication.palette())

    def on_check_box_include_timestamp_state_changed(self, _: Qt.CheckState) -> None:
        self.clear_sign_result()
        self.update_signature_filename()

    def clear_sign_result(self) -> None:
        self.textBrowserResultSign.clear()
        self.textBrowserResultSign.setPalette(QGuiApplication.palette())

    def status_palette(self, status: ResultStatus) -> QPalette:
        if status == ResultStatus.Normal:
            return self.success_palette
        elif status == ResultStatus.Warning:
            return self.warning_palette
        else:
            return self.error_palette

    def start_waiting(self, dialog: WaitDialog) -> None:
        self.waiting_dialog = dialog
        self.waiting_dialog.setWindowTitle(translate('text', 'Processing'))
        self.disable_controls()
        self.waiting_timer.start()

    def stop_waiting(self) -> None:
        self.waiting_timer.stop()
        self.close_waiting_dialog()
        self.enable_controls()

    def show_waiting_dialog(self) -> None:
        if not self.is_cryptopro_ui_running() and self.waiting_dialog:
            self.waiting_dialog.exec()

    def close_waiting_dialog(self) -> None:
        if self.waiting_dialog:
            self.waiting_dialog.close()
        self.waiting_dialog = None

    def show_help(self) -> None:
        url = 'qthelp://flyhelp.ru/doc/fly-csp-cryptopro/fly-csp-cryptopro.html'
        path = '/usr/share/doc/fly/qthelp/ru/flyhelp.qhc'
        QProcess.startDetached("assistant -collectionFile {path} -showUrl {url}".format(path=path, url=url))

    def show_error(self, title: str, error: str, details: str = None, stop_waiting: bool = True) -> None:
        try:
            msgbox = QMessageBox(self)
            msgbox.setWindowTitle(title)
            msgbox.setText(error)
            if details:
                msgbox.setDetailedText(details)
            msgbox.exec()

            if stop_waiting:
                self.stop_waiting()
        except Exception as exception:
            QMessageBox.critical(self,
                                 translate('text', 'Failed show error'),
                                 str(exception))

    def show_critical_error(self, title: str, error: str, details: str = None) -> None:
        self.show_error(title, error, details)
        self.emergency_exit()

    def ask_about_conversion(self) -> bool:
        return QMessageBox.question(self, translate('text', 'Warning'),
                                    translate('text', 'File size exceeds 10 megabytes. '
                                                      'Conversion to PDF format may take a long time. '
                                                      'Do you want to continue?')) == QMessageBox.Yes

    def clear_cache(self) -> None:
        self.cache.clear_cache()

    def emergency_exit(self) -> None:
        self.clear_cache()
        sys.exit(0)

    @staticmethod
    def is_cryptopro_ui_running() -> bool:
        uid = os.getuid()
        for proc in psutil.process_iter():
            try:
                if proc.name() == "xcpui_app" and proc.uids().real == uid:
                    return True
            except psutil.Error:
                pass
        return False

    def setup_cryptopro_bin_dir(self) -> None:
        arch = machine()
        try:
            paths = self.global_settings.cryptopro_bin_paths[arch]
        except KeyError:
            raise Exception(translate('text', 'CryptoPro bin directory not set in settings for {arch} architecture. '
                                              'Please, check /etc/astra-csp-cryptopro/fly-csp-cryptopro.conf')
                            .format(arch=arch))

        for path in paths:
            directory = QDir(path)
            if directory.exists():
                self.cryptopro_bin_dir = directory
                break
        else:
            raise Exception(translate('text', 'CryptoPro bin directory {paths} not exist. '
                                              'Please, check /etc/astra-csp-cryptopro/fly-csp-cryptopro.conf')
                            .format(paths=', '.join(paths)))

    def validate_system(self) -> None:
        try:
            current_max_level = Utils.get_mac_level()
        except:
            raise Exception(translate('text', 'Error getting mandated level'))

        available_max_level = self.global_settings.max_mac_level
        if current_max_level > available_max_level:
            message = translate('text', 'The use of csp service at mandated '
                                        'level {current}(>{available}) is prohibited').format(current=current_max_level,
                                                                                              available=available_max_level)
            raise Exception(message)

        if not Utils.file_is_exists(self.cryptopro_bin_dir.filePath('cryptcp')):
            raise Exception(translate('text', 'Crypto provider not installed'))

        if not Utils.file_is_exists(self.cryptopro_bin_dir.filePath('certmgr')):
            raise Exception(translate('text', 'Certificate manager not installed'))

    def unify_buttons_sizes(self):
        buttons = [self.pushButtonSelectContainer,
                   self.pushButtonSelectContainerSign,
                   self.pushButtonSelectDocument,
                   self.pushButtonSelectDocumentSign,
                   self.pushButtonDocumentView,
                   self.pushButtonDocumentViewSign,
                   self.pushButtonSelectSignature,
                   self.pushButtonVerifySignature,
                   self.pushButtonGenerateName,
                   self.pushButtonSign]
        max_width = max([button.width() for button in buttons])
        for button in buttons:
            button.setMinimumWidth(max_width)

    @staticmethod
    def time_mark_plugin_installed() -> bool:
        return Package.installed('cprocsp-pki-cades') or Package.installed('cprocsp-pki-cades-64')
