#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Optional

from PyQt5.QtCore import QProcess, QTimer, QByteArray, pyqtSignal


class CompletedResult:
    def __init__(self,
                 error: QProcess.ProcessError = None,
                 exit_status: QProcess.ExitStatus = None,
                 exit_code: int = None) -> None:
        self.error = error
        self.exit_status = exit_status
        self.exit_code = exit_code


class Process(QProcess):
    completed = pyqtSignal(CompletedResult)

    def __init__(self) -> None:
        super().__init__()
        self.__completed_result = None  # type: Optional[CompletedResult]
        self.__timer = QTimer()
        self.__last_standard_output = QByteArray()
        self.__previous_standard_output = QByteArray()
        self.__setup()

    @property
    def kill_time(self) -> int:
        return self.__timer.interval()

    @kill_time.setter
    def kill_time(self, ms: int) -> None:
        self.__timer.setInterval(ms)

    @property
    def command(self) -> str:
        return ' '.join([self.program()] + self.arguments())

    @property
    def last_standard_output(self) -> QByteArray:
        return self.__last_standard_output

    @property
    def standard_output_buffer(self) -> QByteArray:
        return QByteArray(self.__previous_standard_output).append(self.__last_standard_output)

    @property
    def completed_result(self) -> Optional[CompletedResult]:
        return self.__completed_result

    def clear_standard_output_buffer(self) -> None:
        self.__last_standard_output.clear()
        self.__previous_standard_output.clear()

    def __setup(self) -> None:
        self.__timer.setSingleShot(True)
        self.__timer.timeout.connect(self.__timer_on_timeout)
        self.readyReadStandardOutput.connect(self.__on_ready_read_standard_output)
        self.stateChanged.connect(self.__on_state_changed)
        self.finished.connect(self.__on_finished)
        self.errorOccurred.connect(self.__on_error_occurred)

    def __timer_on_timeout(self) -> None:
        self.kill()

    def __on_starting(self) -> None:
        self.__completed_result = None
        self.clear_standard_output_buffer()
        if self.kill_time > 0:
            self.__timer.start()

    def __on_ready_read_standard_output(self) -> None:
        self.__previous_standard_output += self.__last_standard_output
        self.__last_standard_output = self.readAllStandardOutput()

    def __on_state_changed(self, state: QProcess.ProcessState) -> None:
        if state == QProcess.Starting:
            self.__on_starting()

    def __on_finished(self, exit_code: int, exit_status: QProcess.ExitStatus) -> None:
        self.__timer.stop()

        self.__completed_result = CompletedResult(None, exit_status, exit_code)
        self.completed.emit(self.__completed_result)

    def __on_error_occurred(self, error: QProcess.ProcessError) -> None:
        self.__timer.stop()

        if error == QProcess.Crashed:
            return

        self.__completed_result = CompletedResult(error)
        self.completed.emit(self.__completed_result)
