#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

from PyQt5.QtCore import Qt, QTextStream, pyqtSignal
from PyQt5.QtNetwork import QLocalSocket, QLocalServer
from PyQt5.QtWidgets import QApplication, QMainWindow


class QtSingleApplication(QApplication):
    messageReceived = pyqtSignal(str)

    def __init__(self, _id: str, *argv: List[str]) -> None:
        super(QtSingleApplication, self).__init__(*argv)
        self.__id = _id
        self.__out_socket = QLocalSocket()
        self.__out_socket.connectToServer(self.__id)
        self.__is_running = self.__out_socket.waitForConnected()
        self.__activation_window = None
        self.__out_stream = None
        self.__in_socket = None
        self.__in_stream = None
        self.__server = None
        if self.__is_running:
            self.__out_stream = QTextStream(self.__out_socket)
            self.__out_stream.setCodec('UTF-8')
        else:
            error = self.__out_socket.error()
            if error == QLocalSocket.ConnectionRefusedError:
                self.close()
                QLocalServer.removeServer(self.__id)
            self.__out_socket = None
            self.__server = QLocalServer()
            self.__server.listen(self.__id)
            self.__server.newConnection.connect(self.__on_new_connection)

    @property
    def id(self) -> str:
        return self.__id

    @property
    def is_running(self) -> bool:
        return self.__is_running

    @property
    def activation_window(self) -> QMainWindow:
        return self.__activation_window

    @activation_window.setter
    def activation_window(self, window: QMainWindow) -> None:
        self.__activation_window = window

    def activate_window(self) -> None:
        if not self.__activation_window:
            return
        self.__activation_window.setWindowState(
            self.__activation_window.windowState() & ~Qt.WindowMinimized)
        self.__activation_window.raise_()
        self.__activation_window.activateWindow()

    def close(self) -> None:
        if self.__in_socket:
            self.__in_socket.disconnectFromServer()
        if self.__out_socket:
            self.__out_socket.disconnectFromServer()
        if self.__server:
            self.__server.close()

    def sendMessage(self, message: str) -> bool:
        if not self.__out_stream:
            return False
        self.__out_stream << message << '\n'
        self.__out_stream.flush()
        return self.__out_socket.waitForBytesWritten()

    def __on_new_connection(self) -> None:
        if self.__in_socket:
            self.__in_socket.readyRead.disconnect(self.__on_ready_read)
        self.__in_socket = self.__server.nextPendingConnection()
        if not self.__in_socket:
            return
        self.__in_stream = QTextStream(self.__in_socket)
        self.__in_stream.setCodec('UTF-8')
        self.__in_socket.readyRead.connect(self.__on_ready_read)
        self.activate_window()

    def __on_ready_read(self) -> None:
        while True:
            msg = self.__in_stream.readLine()
            if not msg:
                break
            self.messageReceived.emit(msg)
