#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import shlex
import subprocess
from typing import List

from PyQt5.Qt import QApplication
from PyQt5.QtCore import QProcess

from astra_csp_cryptopro.process import Process
from astra_csp_cryptopro.utils import Utils

translate = QApplication.translate


class FileInfo:
    def __init__(self,
                 length: int,
                 date: str,
                 time: str,
                 name: str) -> None:
        self.length = length
        self.date = date
        self.time = time
        self.name = name

    def __repr__(self) -> str:
        return ' '.join([str(self.length), self.date, self.time, self.name])


class ZipInfo:
    def __init__(self, files_info: List[FileInfo]) -> None:
        self.files_info = files_info

    def __repr__(self) -> str:
        return 'files: ' + ', '.join(file.name for file in self.files_info)

    def file_exists(self, filename: str) -> bool:
        for file_info in self.files_info:
            if file_info.name == filename:
                return True
        else:
            return False


class ZipArchiver:
    def __init__(self, filepath: str, targetdir: str) -> None:
        self.__filepath = filepath
        self.__target = targetdir

    def create(self, update: bool = False) -> None:
        if update:
            args = ['-q', '--junk-paths', '-0', '--compression-method', 'store', '-r', self.__filepath, self.__target]
        else:
            args = ['-q', '--junk-paths', '-0', '--compression-method', 'store', '-FSr', self.__filepath, self.__target]
        Utils.execute('zip', args)

    def unpack(self, files: List[str] = None) -> None:
        if files:
            shielded_filepath = shlex.quote(self.__filepath)
            shielded_target = shlex.quote(self.__target)
            shielded_files = [shlex.quote(path) for path in files]
            for file in shielded_files:
                args = ['unzip', '-pI', 'utf-8', shielded_filepath, file, '>', shielded_target + '/' + file]
                Utils.execute('bash', ['-c', ' '.join(args)])
        else:
            args = ['-qqoI', 'utf-8', self.__filepath, '-d', self.__target]
            Utils.execute('unzip', args)

    def update(self) -> None:
        self.create()

    def info(self) -> ZipInfo:
        process = Process()
        process.start('unzip -lqq "{path}"'.format(path=self.__filepath))
        process.waitForFinished()

        if process.completed_result.error is not None:
            message = translate('text', 'Process error') + '. ' + process.errorString()
            raise Exception(message)
        if process.completed_result.exit_status is not None \
                and process.completed_result.exit_status != QProcess.NormalExit:
            message = translate('text', 'Process error') + '. ' + process.errorString()
            raise Exception(message)
        if process.completed_result is not None and process.completed_result.exit_code != 0:
            message = translate('text', 'Process exited with unexpected code: {code}') \
                .format(code=process.completed_result.exit_code)
            raise Exception(message)

        files_info = []
        for line in Utils.to_utf8(process.standard_output_buffer).split('\n'):
            match = re.match(r'^\s*(?P<length>\d+)\s+(?P<date>\S+)\s+(?P<time>\S+)\s+(?P<name>.+)', line)
            if match:
                file_info = FileInfo(length=int(match.group('length')),
                                     date=match.group('date'),
                                     time=match.group('time'),
                                     name=match.group('name'))
                files_info.append(file_info)

        return ZipInfo(files_info)

    def validate_unpacked_files(self, file_paths: List[str] = None) -> bool:
        info = self.info()

        if file_paths:
            files_paths_for_check = file_paths
        else:
            files_paths_for_check = []
            for filename in [file_info.name for file_info in info.files_info]:
                files_paths_for_check.append(self.__target + '/' + filename)

        for file_path in files_paths_for_check:
            if not Utils.file_is_exists(file_path):
                return False

            filename = os.path.basename(file_path)
            files_info = [file for file in info.files_info if file.name == filename]
            if not files_info:
                return False

            if files_info[0].length != Utils.file_size(file_path):
                return False
        else:
            return True

    @classmethod
    def is_encrypted(csl, filepath: str) -> bool:
        result = subprocess.run('/usr/bin/zipinfo "{filepath}"'.format(filepath=filepath),
                                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                                shell=True, universal_newlines=True, check=True)
        output_re = re.compile('^.*?\\s+.*?\\s+.*?\\s+.*?\\s+(?P<type>[BTbt]).\\s+.*?$', re.MULTILINE | re.U)
        for match in output_re.finditer(result.stdout):
            if match:
                type = match.group('type')
                if type in ['T', 'B']:
                    return True
        return False
